#include <iostream>
#include <list>
#include <string>
#include <algorithm>

#include "arccli.h"

#include <arc/jobftpcontrol.h>
#include <arc/joblist.h>
#include <arc/mdsquery.h>
#include <arc/notify.h>
#include <arc/resource.h>

#ifdef HAVE_LIBINTL_H
#include <libintl.h>
#define _(A) dgettext("arccli", (A))
#else
#define _(A) (A)
#endif


void arcclean (const std::list<std::string>& jobs,
               const std::list<std::string>& clusterselect,
               const std::list<std::string>& clusterreject,
               const std::list<std::string>& status,
               const bool force,
               const int timeout,
               const bool anonymous) {

	std::list<std::string> jobids =
		GetJobIDsList(jobs, clusterselect, clusterreject);

	if (jobs.empty() && clusterselect.empty() && jobids.empty())
		throw ARCCLIError(_("No jobs"));

	if (jobids.empty())
		throw ARCCLIError(_("No valid jobnames/jobids given"));

	std::list<Job> joblist =
		GetJobInfo(jobids, MDS_FILTER_JOBINFO, anonymous, "", timeout);

	JobFTPControl ctrl;

	SortByHost(joblist);

	for (std::list<Job>::iterator jli = joblist.begin();
	     jli != joblist.end(); jli++) {

		if (force && jli->status.empty() && status.empty()) {
			notify(WARNING) << _("Deleting job (from local list only)")
			                << ": " << jli->id << std::endl;
			RemoveJobID(jli->id);
			continue;
		}

		if (jli->status.empty()) {
			notify(WARNING) << _("Job information not found")
			                << ": " << jli->id << std::endl;
			continue;
		}

		if (!status.empty() &&
		    std::find(status.begin(), status.end(), jli->status) == status.end())
			continue;

		if (jli->status != "FINISHED" && jli->status != "FAILED" &&
		    jli->status != "KILLED" && jli->status != "DELETED") {
			notify(WARNING) << _("Job has not finished yet")
			                << ": " << jli->id << std::endl;
			continue;
		}

		notify(WARNING) << _("Deleting job")
		                << ": " << jli->id << std::endl;

		try {
			ctrl.Clean(jli->id, timeout, false);
		} catch(ARCLibError& e) {
			notify(WARNING) << e.what() << std::endl;
			continue;
		}

		RemoveJobID(jli->id);
	}
}
