/*
 * Created on Sep 27, 2005
 *
 *
 */
package se.umu.jobsubmission.impl.dispatcher;

import java.io.File;
import java.net.URI;
import java.net.URL;
import java.security.cert.X509Certificate;
import java.util.Calendar;

import javax.xml.rpc.Stub;

import org.apache.axis.components.uuid.UUIDGenFactory;
import org.apache.axis.components.uuid.UUIDGen;
import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.AttributedURI;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.PositiveInteger;
import org.apache.log4j.Logger;
import org.globus.delegation.DelegationConstants;
import org.globus.delegation.DelegationUtil;
import org.globus.exec.generated.CreateManagedJobInputType;
import org.globus.exec.generated.CreateManagedJobOutputType;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.ManagedJobFactoryPortType;
import org.globus.exec.utils.ManagedJobFactoryConstants;
import org.globus.exec.utils.client.ManagedJobFactoryClientHelper;
import org.globus.exec.utils.rsl.RSLHelper;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.globus.mds.glue.GLUECERPType;
import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.TransferType;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.HostAuthorization;
import org.globus.wsrf.impl.security.descriptor.ClientSecurityDescriptor;
import org.ietf.jgss.GSSCredential;

import se.umu.jobsubmission.util.JobDescriptionTranslator;
import se.umu.wsag.util.JobSubmissionUtils;

/**
 * @author tordsson
 *
 * Basic dispatcher for GT4 (WS-GRAM). 
 * The descriptiontranslator needs to be completed. 
 */
public class GT4Dispatcher implements Dispatcher {
	
	private static Logger logger = Logger.getLogger(GT4Dispatcher.class); 
	private JobDescriptionTranslator descriptionTranslator; 
	
	public GT4Dispatcher() {
		descriptionTranslator = (JobDescriptionTranslator) new JSDL2RSL(); 
	}
	
	/**
	 * Creates URL to ManagedJobFactoryService, 
	 * sets client security descriptor, 
	 * delegates credential, creates job resource
	 * and hence submits job. 
	 */
	public Object SubmitJob(JobDescriptionType jobDescription, 
			GLUECERPType resourceInfo, 
			boolean clientStagesInput) throws Exception {
		
		// [protocol://]host[:port][/servicepath]
		
        String resourceHost = resourceInfo.getCluster(0).
			getSubCluster(0).getHost(0).getName(); 
        
        // Cheated a bit just to test...-> it worked
        // resourceHost = "coppi.cs.umu.se"; 
		
		// 3. Resolve factory URL
		URL factoryUrl = ManagedJobFactoryClientHelper.getServiceURL(
				resourceHost).getURL();
		
		logger.debug("Dispatching job to: " + factoryUrl); 
		
        // factoryType = ManagedJobFactoryConstants.FACTORY_TYPE.PBS; 
		
		// TODO: Check in the glue-info which batch system type that
		// is used (PBS, LSF, Condor) 
        EndpointReferenceType factoryEndpoint = ManagedJobFactoryClientHelper.
			getFactoryEndpoint(factoryUrl, 
					ManagedJobFactoryConstants.FACTORY_TYPE.PBS); 

		ManagedJobFactoryPortType factoryPort = ManagedJobFactoryClientHelper.
				getPort(factoryEndpoint);

		// 4. Set client security properties.
		ClientSecurityDescriptor secDesc = new ClientSecurityDescriptor();
		// secDesc.setGSITransport(Constants.<protection level constant>);
		secDesc.setGSITransport(Constants.ENCRYPTION); // Really guessing here... 
		//secDesc.setAuthz(<Authorization sub-class instance>);
		secDesc.setAuthz(HostAuthorization.getInstance()); 
		
		// GlobusCredential globusCredential = GlobusCredential.getDefaultCredential(); 
		// Get delegated credential, and delegate it further. 
		
		GlobusCredential globusCredential = JobSubmissionUtils.getDelegatedCredential(); 
		
		GSSCredential credential = new GlobusGSSCredentialImpl(globusCredential, 
				GSSCredential.INITIATE_AND_ACCEPT); 
		if(credential != null) {
			secDesc.setGSSCredential(credential); 
		}
		
		((Stub) factoryPort)._setProperty(Constants.CLIENT_DESCRIPTOR, secDesc);
		
        // Delegate Credentials
		// -> Not needed until we do file staging. 
    	logger.debug("Delegating credentials..."); 
    	
		String URL = ServiceHost.getDefaultProtocol() + "://" // "https://" 
			         + resourceHost 
					 + ":" + ServiceHost.getDefaultPort()   // ":8443"
					 + DelegationConstants.SERVICE_BASE_PATH // "wsrf/services/" 
					 + DelegationConstants.FACTORY_PATH; // + "DelegationFactoryService"; 		
		
		EndpointReferenceType delegationEpr = null; 
		logger.debug("Delegating credential to: " + URL); 
		
	   	EndpointReferenceType DelegationServiceEPR = new EndpointReferenceType(); 
    	DelegationServiceEPR.setAddress(new Address(URL)); 
    	X509Certificate[] certs = DelegationUtil.getCertificateChainRP(DelegationServiceEPR,
    			secDesc);    	
    	
    	delegationEpr = DelegationUtil.delegate(
    			URL, 
				globusCredential, 
				certs[0], 
				true, 
				secDesc); 
		
		// Credential to execute job
		jobDescription.setJobCredentialEndpoint(delegationEpr);
    	
		// Credential for stage in 
		if(!clientStagesInput) {
			jobDescription.setStagingCredentialEndpoint(delegationEpr);
			TransferRequestType stageIn = jobDescription.getFileStageIn(); 
	        if(stageIn != null) {
	        	logger.debug("Setting delegated credential for job stage in"); 
	        	stageIn.setTransferCredentialEndpoint(delegationEpr); 
	        }
	        else {
	        	logger.debug("Job stage in does not require delegated credentials"); 
	        }
		}
		
		// Credential for stage out. 
		TransferRequestType stageOut = jobDescription.getFileStageOut();
		if(stageOut != null) {
			logger.debug("Setting Delegated credential for job stage out"); 
			stageOut.setTransferCredentialEndpoint(delegationEpr);
		}
		else {
			logger.debug("Job stage out does not require delegated credentials"); 
		}
		
        // 7. Creating the job resource.
		CreateManagedJobInputType jobInput = new CreateManagedJobInputType();
		// N.B: This was incorrectly described at 
		UUIDGen uuidGenFactory = UUIDGenFactory.getUUIDGen(); 
		String uuidGen = uuidGenFactory.nextUUID();		
		jobInput.setJobID(new AttributedURI("uuid:" + uuidGen));
		
		
		// Set job (job resource) termination time
		// TODO: This must be set to a value at least as high
		// as the job run-time, probably much higher due to
		// batch queue waiting. 
		
		// What happens when this WS-Resource is destroyed, is
		// job cancelled or not?
		Calendar cal = Calendar.getInstance(); 
		cal.add(Calendar.HOUR, 24); 		
		jobInput.setInitialTerminationTime(cal);
		jobInput.setJob(jobDescription);
		
		// Send job request to MJFS
		logger.debug("Creating job resource..."); 
		CreateManagedJobOutputType createResponse = factoryPort.
			createManagedJob(jobInput);
		EndpointReferenceType jobEndpoint = createResponse.getManagedJobEndpoint();
		
		// Serialize and return job endpoint
		String jobId = ObjectSerializer.toString(jobEndpoint, 
				org.apache.axis.message.addressing.Constants.QNAME_ENDPOINT_REFERENCE); 
		logger.debug("Job submitted with id: " + jobId); 
		return jobId; 
		
	}
	
	/**
	 * Prepares the rsl for job submission. 
	 * Sets some undefined attributes, add reservation to rsl, 
	 * and modify input files to ensure correct staging if 
	 * the JSS and the client are run on different hosts. 
	 * @param rsl
	 * @param reservation
	 * @param clientURL
	 */
	public void fixRSL(JobDescriptionType rsl, 
			Object reservation, 
			String clientURL) {
		
		// Set cputime and count to default values unless they 
    	// are specified by the user. 
    	if(rsl.getCount() == null) {
    		PositiveInteger count = new PositiveInteger(""
    				+ Dispatcher.DEFAULT_CPU_COUNT); 
    		rsl.setCount(count); 
    	}
    	if(rsl.getMaxCpuTime() == null) {
    		Long cpuTime = new Long((long) Dispatcher.DEFAULT_CPU_TIME); 
    		rsl.setMaxCpuTime(cpuTime); 
    	}
    	
    	// Add advance reservation 
    	if(reservation != null) {
    		
    	}
		
	}
	
	/**
	 * Ensure that local input files gets staged by 3rd party transfer. 
	 * This is only used if the client does not stage local input files. 
	 * @param rsl
	 * @param clientURL
	 */
	public void fixInputFiles(JobDescriptionType rsl, String clientURL) {
    	
    	// If any input file starts with file:///fileName, replace this 
    	// with gsiftp://clientURL:2811/fileName
    	TransferRequestType stageIn = rsl.getFileStageIn(); 
    	if(stageIn != null) {
    		TransferType transfers[] = stageIn.getTransfer(); 
    		for(int i = 0; i < transfers.length; i++) {
    			String source = transfers[i].getSourceUrl(); 
    			try {
    				URI uri = new URI(source); 
    				if("file".equals(uri.getScheme())) {
    					logger.debug("Changing input file URL");
    					String path = uri.getSchemeSpecificPart(); 
    					File f = new File(path); 
    			    	String newURL = "gsiftp://" 
    			    				+ clientURL
    								+ ":2811" 
    								+ f.getAbsolutePath(); 
    			    	transfers[i].setSourceUrl(newURL); 
    				}
    			} catch(Exception e) {
    				logger.error("Incorrect specified input files URL" 
    						+ source); 
    			}
    		}
    	}
	}
	
    public Object requestJob(Object resource,   
            Object jobDescription, 
            Object reservation, 
            String clientURL, 
			boolean clientStagesInput) throws Exception {
    	
    	// Throws Exception unless 
    	logger.info("Dispatching job using: " + GT4Dispatcher.class); 
    	
    	GLUECERPType glueInfo = (GLUECERPType) resource; 
    	
    	logger.debug("Resource name is: " + 
    			glueInfo.getCluster(0).getSubCluster(0).getHost(0).getName()); 
    	
    	String resourceURL = ((GLUECERPType)resource).getCluster(0).
			getSubCluster(0).getHost(0).getName(); 
    	
    	// First ensure that selected resource runs a service container.
    	// As call to WS is slow, this saves a lot of time if the resource
    	// does not run a service container. 
    	if(!JobSubmissionUtils.ping(resourceURL)) {
    		logger.warn("No service container found at: " 
    				+ resourceURL + " job submission failed"); 
    		throw new Exception("No service container found at: " 
    				+ resourceURL + " job submission failed"); 
    	}
  
    	JobDescriptionType rsl = (JobDescriptionType) 
			this.descriptionTranslator.translate(jobDescription); 
    	
    	logger.debug("Rsl is: "); 
    	logger.debug(RSLHelper.convertToString(rsl)); 
    	
    	fixRSL(rsl, reservation, clientURL); 
    	
    	if(!clientStagesInput) {
    		logger.info("Preparing staging of local input files"); 
    		fixInputFiles(rsl, clientURL); 
    	}
    	else {
    		logger.info("No staging performed (it is handled by client)"); 
    	}
    	
    	logger.debug("Final rsl is: "); 
    	logger.debug(RSLHelper.convertToString(rsl)); 
    	
    	Object result = this.SubmitJob(rsl, 
    			(GLUECERPType) resource, 
				clientStagesInput); 
    	
    	return result; 
    	
    }
}
