/*
 * Created on Feb 3, 2006
 *
 */
package se.umu.jobsubmission.util;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.globus.rsl.AbstractRslNode;
import org.globus.rsl.NameOpValue;
import org.globus.rsl.Value;

import org.apache.log4j.Logger;

/**
 * @author tordsson
 *
 * Contains some code for xRSL manipulation used by both 
 * the ARC client plugin and the ARCDispatcher
 */
public class XRSLUtils {

    private static Logger logger = Logger.getLogger(XRSLUtils.class); 

    /*
       For an xrsl, the value of the relation 
       (stdout=myjob.txt) 
       shout be added to the relation (outputfiles=( ... )), 
       unless there already exists an entry for stdout in outputfiles 

       TODO: Perform some typechecking of values, to ensure that
       they are of correct object types, see xrsl2jsdl.java for details. 
*/
    public static void addIOStaging(AbstractRslNode rsl, 
            String fileRelationName, 
            String fileListRelationName) throws Exception {

        // Check if relation was defined
        NameOpValue fileRelation = rsl.getParam(fileRelationName); 
        if(fileRelation == null) 
            return; 

        // Get value of fileRelation
        List allFileRelationValues = fileRelation.getValuesAsStrings(true); 
        String fileRelationValue = (String) allFileRelationValues.get(0); 

        logger.debug("Adding " + fileRelationValue + " to relation " 
                + fileListRelationName); 

        // Check if fileListRelation exists. 
        NameOpValue fileListRelation = rsl.getParam(fileListRelationName); 

        // Add fileListRelation to rsl, and add fileRelation to 
        // fileListRelation
        if(fileListRelation == null) {
            logger.debug("Adding new relation, " + fileListRelationName); 
            List newValues = new LinkedList(); 
            // Name "" is just a dummy, but ensures compliance with NG UI
            newValues.add(new Value(fileRelationValue)); 
            newValues.add(new Value("")); 
            NameOpValue newRelation = 
                new NameOpValue(fileListRelationName, NameOpValue.EQ); 
            newRelation.add(newValues); 
            rsl.add(newRelation); 
            return; 
        }

        // fileListRelation did exist, check if it contains
        // entry for the value of fileRelation
        List fileListRelationValues = 
            fileListRelation.getValuesAsStrings(true); 
        Iterator iter = fileListRelationValues.iterator(); 
        while(iter.hasNext()) { 
            List pair; 
            try {
                pair = (List) iter.next(); 
            } catch (ClassCastException e) {
                throw new Exception("Relation " + fileListRelationName 
                        + " must be a list of lists"); 
            }
            String localName = (String) pair.get(0); 
            if(localName.equals(fileRelationValue)) 
                return; 
        }

        // Entry for fileRelation is missing in fileListRelation, add it
        List newEntry = new LinkedList(); 
        // The name "" ensures that file is transferred back to used host by UI
        newEntry.add(new Value(fileRelationValue)); 
        newEntry.add(new Value("")); 
        fileListRelation.add(newEntry); 

    }    

}
