
package se.umu.jobsubmission.impl.informationfinder; 
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;

import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;

import org.apache.log4j.Logger;
import org.globus.gsi.GlobusCredential;

import se.umu.wsag.util.JobSubmissionUtils;

public class ARCInformationFetcher implements InformationFetcher {

	private static Logger logger = Logger.getLogger(ARCInformationFetcher.class); 
	
    private List resources;     

    public ARCInformationFetcher() {
        resources = new LinkedList(); 
    }

    public int size() {
        return resources.size();
    }
    
    public String translateURL(String URL) {
    	
    	if(URL.startsWith("ldap://"))
    		return URL; 
    	else
    		return "ldap://" + URL + ":2135/nordugrid-cluster-name=" + URL 
			+ ",Mds-Vo-name=local,o=grid"; 
    }
    
    /** resourceURI can either have format: 
     * "ldap://hostName:2135/nordugrid-cluster-name=hostName,
     *        Mds-Vo-name=local,o=grid"
     * or format "hostName"
     */
    public Object getInformation(String resourceURL, int timeout) 
    	throws NamingException {
    	
        ARCResource resource = null; 
        DirContext ctx = null; 
        
        Attributes cluster = null;
        Attributes queue = null;
        Attributes authuser = null; 
        String validTo = null; 
        
        // Translate resourceURI to ldap://... unless it already has this format.
        String ldapURL = translateURL(resourceURL); 
        
        // TODO: Broken!
        // This can not be used if user specifies submission target in 
        // job preferences file! 
        // This test ensures that the address is directed to a ARC ldap server, 
        // not a lcg one. 
        /* 
        if(!ldapURL.matches(".*[Nn]ordugrid-[Cc]luster-[Nn]ame.*")) {
        	logger.warn("Non-ARC LDAP server URL specified: "+ ldapURL); 
        	throw new NamingException("Non-ARC LDAP server URL specified: " 
        			+ resourceURL); 
        }
        */
        
        logger.debug("ResourceURL: " + resourceURL); 
        logger.debug("ldapURL: " + ldapURL); 

        // Run query
        try {
            Hashtable env = new Hashtable();
            env.put(Context.INITIAL_CONTEXT_FACTORY,
                          "com.sun.jndi.ldap.LdapCtxFactory");
            env.put(Context.PROVIDER_URL, ldapURL);
            env.put("com.sun.jndi.ldap.connect.timeout", ""+timeout);
            ctx = new InitialDirContext(env);
            SearchControls controls = new SearchControls();
            controls.setCountLimit(0); // No limit, return all entries
            controls.setTimeLimit(timeout);
            controls.setSearchScope(SearchControls.SUBTREE_SCOPE);
            
            String DN = null; 
            try {
            	GlobusCredential cred = JobSubmissionUtils.getDelegatedCredential();             	
                DN = cred.getIdentity();                    
            } catch (Exception eer) {}
            String filter = 
                 "(|"+
                   "(objectclass=nordugrid-cluster)"+
                   "(&"+ 
				   "(objectclass=nordugrid-queue)" + 
				   "(nordugrid-queue-status=active)"+
				   ")"+
		           "(&"+
                   "(objectclass=nordugrid-authuser)"+
                   "(nordugrid-authuser-sn="+DN+")"+
                   ")"+
                 ")";
            
            NamingEnumeration resultp = ctx.search("",filter,controls);
            /*
            Attributes cluster = null;
            Attributes queue = null;
            Attributes authuser = null; 
            String validTo = null; 
            */
            
            if(!resultp.hasMore()) {
            	logger.info("Probably querying an lcg resource..."); 
            	throw new NamingException("No reply from resource: " + ldapURL); 
            }
            
            while(resultp.hasMore()) {
                try {
                    SearchResult result = (SearchResult)(resultp.next());
                    Attributes attrs = result.getAttributes();
                    
                    Attribute valid = attrs.get("Mds-validto"); 
                    if(valid != null) {
                    	validTo = (String) valid.get();                
                    }
                    
                    Attribute attr = attrs.get("objectClass");
                    
                    String oclass = null;
                    for(int i=0;i<attrs.size();i++) {
                        oclass=(String)(attrs.get("objectClass").get(i));
                        if(oclass.equals("nordugrid-cluster")) {
                            cluster=attrs; 
                            break;
                        } 
                        else if(oclass.equals("nordugrid-queue")) {
                            queue=attrs; 
                            break;
                        }
                        else if(oclass.equals("nordugrid-authuser")) {
                        	// logger.warn(("Found obj. nordugrid-authuser for " + resourceURI)); 
                            authuser=attrs; 
                            break; 
                        }
                    }
                    /*
                    if(authuser == null) 
                    resource = new ARCResource(cluster, queue, validTo); 
                    */
                } catch(java.lang.Exception err) {
                    logger.error("Failed to fetch information: " 
                    		+ err.toString());
                    continue;
                }                  
            }
        } catch (NamingException err) { 
            logger.error("Failed to fetch information: "
            		+ err.toString());
            throw err; 
        } 
        finally {
            if(ctx != null) {
                try {
                    ctx.close();                   
                } catch (Exception e) {
                	logger.warn("Error shutting down connection"); 
                }
            }
        }
        
        // Only save resource if user is authorized to use it... 
        if(authuser == null) {
        	return null; 
        }
        resource = new ARCResource(cluster, queue, validTo); 

        return resource; 
    }
}
