package se.umu.jobsubmission.impl.informationfinder;

import java.util.List; 
import java.util.Vector; 

import org.apache.log4j.Logger; 
/**
 * @author tordsson
 *
 * Trivial factory for InformationFinders
 */
public class InformationFinderFactory {
	
	private static Logger logger = Logger.getLogger(InformationFinderFactory.
			class); 
	
	private int finderThreads; 
	private String finderClassNames[]; 
	private String fetcherClassNames[]; 
	private String converterClassNames[]; 
	private int timeout; 
	private List indexServersList; 	
	private ResourceInformationCache resourceInfoCache; 
	
	// Create an instance of each specified class to ensure 
	// that no invalid classNames are specified in the config file...
	// Returns an array of the valid classNames, or null if 
	// no class name was valid.  
	public String[] filterList(String classNames[]) {
		
		if(classNames == null) 
			return null; 
		
		List goodOnes = new Vector(); 
		for(int i = 0; i < classNames.length; i++) {
			try {
				logger.debug("Class name: " + classNames[i]); 
				Class.forName(classNames[i]).newInstance(); 
				goodOnes.add(classNames[i]); 
			} catch (Exception e) {
				logger.error("Bad class name: " + classNames[i]); 
			}
		}
		if(goodOnes.size() == 0)
			return null; 
		else {
			String result[] = new String[goodOnes.size()]; 
			for(int i = 0; i < goodOnes.size(); i++) {
				result[i] = (String) goodOnes.get(i); 
				logger.debug("Good class name: " + result[i]); 
			}
			return result; 
		}
	}
	
	public InformationFinderFactory(int finderThreads, 
			String finderClassNames[], 
			String fetcherClassNames[], 
			String converterClassNames[], 
			int timeout, 
			List indexServersList, 
			ResourceInformationCache resourceInfoCache) throws Exception {
		
		// We can still submit jobs to pre-defined targets without 
		// this one, hence logger.error and not logger.fatal...
		this.finderClassNames = filterList(finderClassNames); 
		if(this.finderClassNames == null) {
			logger.error("No valid resource finder specified"); 
			throw new Exception("No valid resource finder specified"); 
		}
		
		this.fetcherClassNames = filterList(fetcherClassNames); 
		if(this.fetcherClassNames == null) {
			logger.fatal("No valid information fetcher specified"); 
			throw new Exception("No valid information fetcher specified"); 
		}
		
		this.converterClassNames = filterList(converterClassNames); 
		/**
		 * This may be null (if Grid middleware uses glue format, 
		 * no conversion is required...) 
		if(this.converterClassNames == null) {
			logger.fatal("No valid information converter specified"); 
			throw new Exception("No valid information converter specified"); 
		}
		*/ 
		
		this.finderThreads = finderThreads; 
		this.timeout = timeout; 
		this.indexServersList = indexServersList; 
		this.resourceInfoCache = resourceInfoCache; 
	}
	
	public InformationFinder getInformationFinder() {
		
		return new InformationFinder(finderThreads, 
				finderClassNames, 
				fetcherClassNames, 
				converterClassNames,
				timeout, 
				indexServersList, 
				resourceInfoCache); 
	}

}
