/*
 * Created on Oct 28, 2005
 *
 * TODO Implement it, then write some documentation
 *
 */
package se.umu.jobsubmission.client;

import java.io.File;
import java.io.FileReader;
import java.net.InetAddress;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import org.apache.axis.message.MessageElement;
import org.apache.axis.types.NCName;
import org.apache.axis.types.NormalizedString;
import org.apache.axis.types.URI;
import org.ggf.schemas.jsdl._2005._06.jsdl.Application_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.CPUArchitecture_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.DataStaging_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.Exact_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.JobDefinition_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.JobDescription_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.OperatingSystemTypeEnumeration;
import org.ggf.schemas.jsdl._2005._06.jsdl.OperatingSystemType_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.OperatingSystem_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.ProcessorArchitectureEnumeration;
import org.ggf.schemas.jsdl._2005._06.jsdl.RangeValue_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.Resources_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.SourceTarget_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl_posix.Argument_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl_posix.Environment_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl_posix.FileName_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl_posix.POSIXApplication_Type;
import org.globus.wsrf.utils.AnyHelper;

import se.umu.jobsubmission.util.JobDescriptionTranslator;
import condor.classad.AttrName;
import condor.classad.ClassAdParser;
import condor.classad.Constant;
import condor.classad.Expr;
import condor.classad.ListExpr;
import condor.classad.Op;
import condor.classad.RecordExpr;
import condor.classad.SelectExpr;

/**
 * @author tordsson
 *
 * Translates EGEE JDL (ClassAds) to JSDL. 
 * The translation of the "Requirements" attribute 
 * is not implemented yet. 
 */
public class JDL2JSDL  implements JobDescriptionTranslator {
	
	/**
	  Translates the free-text "Requirements" attribute to jsdl, 
	  which is rather messy
	  
	   Quote from the lcg manual:
	   "Requirements ... can be arbitrary expressions which use 
	   the fields published by the resources in the MDS system."
	   
	   Translated attributes: 
	   
	   other.GlueHostOperatingSystemName -> jsdl.resource.OperatingSystem
	   other.GlueHostMainMemoryRAMSize	 -> jsdl.resource.individualPhysicalMemory
	   
       These attributes are found in various examples, but are not 
	   defined using their proper Glue names...
	   other.Architecture 		-> jsdl.resource.CPUArchitecture
	   other.OpSys				-> jsdl.resource.OperatingSystem
	   other.MaxCpuTime   		-> jsdl.resource.IndividualCPUTime 
	   
	   
	   other.GlueHostLocalFileSystemSize -> jsdl.resource.individualDiskSpace ** 
	   	   
	   other.GlueCEStateFreeCPUs		 -> ?? 
	   
	   ** Note that the GlueHostLocalFileSystemSize attribute is specified 
	   separately for each file system, /, /tmp, /scratch etc. (and /etc ;-))
	  
	   Not implemented mappings (handled elsewhere): 
	   Preselected hosts are specified in the job preferences document instead:
	   other.GlueHostName				-> jsdl.resource.candidateHosts
	   
	   The NodeNumber attributes specifies the number of CPUs
	   other.GLUECEInfoTotalCPUs		 -> jsdl.resource.totalCPUCount
	  
	 * @param jsdl
	 * @param requirement
	 */
	public void translateRequirements(JobDescription_Type jsdl, 
			Expr requirements) {
		
		// requirements is either of type condor.classad.Op
 		// or of class condor.classad.SelectExpr which inherits Expression. 		
		if(Op.class.equals(requirements.getClass())) {		
			try {
				List allReqs = getRequirements(requirements); 
			
				Iterator iter = allReqs.iterator(); 
				while(iter.hasNext()) {
					translateSingleRequirement(jsdl, (Op)iter.next()); 
				}			
			} catch (Exception e) {
				System.err.println("Invalid \"Requirements\" attribute "
						+ "ignoring it"); 
			}
		}
	}
	
	/**
	 * Translates a single requirement to jsdl
	 * Translated attributes are: 
	 * 		Architecture (GlueSubClusterArchitecture) 
	 * 		OpSys		 (GlueHostOperatingSystemName) 
	 * 		GlueHostLocalFileSystemSize
	 * 		GlueMainMemoryRAMSize
	 * 		
	 * TODO: No operators implemented, always assuming "=="
	 * Operators can be translated to JSDL RangeValueTypes. 
	 * @param jsdl
	 * @param expr
	 */
	public void translateSingleRequirement(JobDescription_Type jsdl, Op expr) {
		
		try {
			String name = ((SelectExpr) expr.arg1).toString(); 
			
			// Translate CPU architecture, there exist some additional 
			// architectures in the enumeration, but these should be the most
			// common ones. 
			// TODO: Could possible use reflection for these. 
			if("other.Architecture".equalsIgnoreCase(name) ||
					"other.GlueSubClusterArchitecture".equalsIgnoreCase(name)) { 
				
				String value = ((Constant) expr.arg2).stringValue(); 					
				CPUArchitecture_Type arch = null; 
				
				if("x86".equalsIgnoreCase(value)) {
					arch = new CPUArchitecture_Type(); 
					arch.setCPUArchitectureName(ProcessorArchitectureEnumeration.x86); 
				}
				// Actually, its called x86 :-) 
				if("intel".equalsIgnoreCase(value)) {
					arch = new CPUArchitecture_Type(); 
					arch.setCPUArchitectureName(ProcessorArchitectureEnumeration.x86); 
				}
				if("ia64".equalsIgnoreCase(value)) {
					arch = new CPUArchitecture_Type(); 
					arch.setCPUArchitectureName(ProcessorArchitectureEnumeration.ia64); 
				}
				// I wonder how this one differs from "x86"? 
				if("x86_32".equalsIgnoreCase(value)) {
					arch = new CPUArchitecture_Type(); 
					arch.setCPUArchitectureName(ProcessorArchitectureEnumeration.x86_32); 
				}
				if("x86_64".equalsIgnoreCase(value)) {
					arch = new CPUArchitecture_Type(); 
					arch.setCPUArchitectureName(ProcessorArchitectureEnumeration.x86_64); 
				}
				if("sparc".equalsIgnoreCase(value)) {
					arch = new CPUArchitecture_Type(); 
					arch.setCPUArchitectureName(ProcessorArchitectureEnumeration.sparc); 
				}
				if("mips".equalsIgnoreCase(value)) {
					arch = new CPUArchitecture_Type(); 
					arch.setCPUArchitectureName(ProcessorArchitectureEnumeration.mips); 
				}
				if("powerpc".equalsIgnoreCase(value)) {
					arch = new CPUArchitecture_Type(); 
					arch.setCPUArchitectureName(ProcessorArchitectureEnumeration.powerpc); 
				}
				
				if(arch != null) {
					jsdl.getResources().setCPUArchitecture(arch); 
				}
				else {
					System.err.println("Unknown Architecture: " + value); 
				}
				
				// Translate some common operating systems name. 
				// This list could be extended with quite a few more. 
				// TODO: Use reflection instead for this one? 
			} else if("other.OpSys".equalsIgnoreCase(name) || 
					"other.GlueHostOperatingSystemName".equalsIgnoreCase(name)) {
				
				String value = ((Constant) expr.arg2).stringValue(); 					
				OperatingSystemType_Type osType = null; 
				
				// Exist quite a few more, but these should cover most cases...
				if("linux".equalsIgnoreCase(value)) {
					osType = new OperatingSystemType_Type(); 
					osType.setOperatingSystemName(OperatingSystemTypeEnumeration.LINUX); 
				}
				if("aix".equalsIgnoreCase(value)) {
					osType = new OperatingSystemType_Type(); 
					osType.setOperatingSystemName(OperatingSystemTypeEnumeration.AIX); 
				}
				if("irix".equalsIgnoreCase(value)) {
					osType = new OperatingSystemType_Type(); 
					osType.setOperatingSystemName(OperatingSystemTypeEnumeration.IRIX); 
				}
				if("Solaris".equalsIgnoreCase(value)) {
					osType = new OperatingSystemType_Type(); 
					osType.setOperatingSystemName(OperatingSystemTypeEnumeration.Solaris); 
				}
				if("SunOS".equalsIgnoreCase(value)) {
					osType = new OperatingSystemType_Type(); 
					osType.setOperatingSystemName(OperatingSystemTypeEnumeration.SunOS); 
				}
				
				if(osType != null) {
					OperatingSystem_Type os = new OperatingSystem_Type(); 
					os.setOperatingSystemType(osType);					
					jsdl.getResources().setOperatingSystem(os); 
				} 
				else {
					System.err.println("Unknown Operating system: " + value); 
				}
			} else if("other.GlueHostLocalFileSystemSize".equalsIgnoreCase(name)) {
				
				int value = ((Constant) expr.arg2).intValue(); 	
				RangeValue_Type valueType = new RangeValue_Type(); 
				Exact_Type exactValue[] = new Exact_Type[1]; 
				exactValue[0] = new Exact_Type(); 
				exactValue[0].set_value(new Double(value).doubleValue()); 
				valueType.setExact(exactValue); 
				jsdl.getResources().setIndividualDiskSpace(valueType); 
				
			} else if("other.GlueHostMainMemoryRAMSize".equalsIgnoreCase(name)) {
				
				int value = ((Constant) expr.arg2).intValue(); 	
				RangeValue_Type valueType = new RangeValue_Type(); 
				Exact_Type exactValue[] = new Exact_Type[1]; 
				exactValue[0] = new Exact_Type(); 
				exactValue[0].set_value(new Double(value).doubleValue()); 
				valueType.setExact(exactValue); 
				// Assuming this is memory per node, not total amount of memory. 
				jsdl.getResources().setIndividualPhysicalMemory(valueType); 
				
			}
			
			else if("other.MaxCpuTime".equalsIgnoreCase(name)) {
				int value = ((Constant) expr.arg2).intValue(); 	
			// TODO: Use this instead of "CPUTime" JDL extension (hack)? 
			// --> Would still need a way to specify walltime
				
			}
			else {
				System.err.println("Unknown requirement: " + name); 
			}
			
			/* This is actually handled by the nodenumber attribute... 
			} else if("other.GLUECEInfoTotalCPUs".equalsIgnoreCase(name)) {
				Resources_Type resource = jsdl.getResources();
				RangeValue_Type valueType = new RangeValue_Type(); 
				Exact_Type exactValue[] = new Exact_Type[1]; 
				exactValue[0] = new Exact_Type(); 
				exactValue[0].set_value(new Double(value).doubleValue()); 
				valueType.setExact(exactValue); 
				resource.setTotalCPUCount(valueType); 
			} */ 
			
		} catch (Exception e) {
			System.err.println("Ignoring malformed requirement: " + expr); 
		}
	}

	
	/**
	 * Recursivly parse requirements expression, split 
	 * @param reqs
	 * @return
	 */
	public List getRequirements(Expr reqs) {
		
		List result = new LinkedList(); 
		Op op = (Op) reqs; 
		
		// If (left) child is not an Op, stop. 
		if(!(Op.class.equals(op.arg1.getClass()))) {
			result.add(op); 
			return result; 
		}
		List left = getRequirements(op.arg1); 
		List right = getRequirements(op.arg2); 
		left.addAll(right); 
		return left; 
	}
	
	/**
	 * Adds a datatransfer to the datastating element 
	 * of the jsdl document. 
	 * @param jsdl
	 * @param newTransfer
	 */
	public void addTransfer(JobDescription_Type jsdl, 
			DataStaging_Type newTransfer) {
		
		DataStaging_Type dataStaging[] = jsdl.getDataStaging(); 
		
		// Add a new DataStaging, or if one already exists, 
		// replace it with one containing the new transfer as well. 
		if(dataStaging == null) {
			dataStaging = new DataStaging_Type[]{newTransfer}; 						
			jsdl.setDataStaging(dataStaging);			
		}
		else {
			int size = dataStaging.length; 
			DataStaging_Type newDataStaging[] = 
				new DataStaging_Type[size + 1]; 
			for(int j = 0; j < size; j++) { 
				newDataStaging[j] = new DataStaging_Type(
						dataStaging[j].getCreationFlag(), 
						dataStaging[j].getDeleteOnTermination(),
						dataStaging[j].getFileName(), 
						dataStaging[j].getFilesystemName(), 
						dataStaging[j].getName(), 
						dataStaging[j].getSource(),
						dataStaging[j].getTarget()); 
			}
			newDataStaging[size] = newTransfer; 
			jsdl.setDataStaging(newDataStaging); 
		}
	}
	
	/**
	 * Translates one attribute of the JDL job description to 
	 * JSDL. The translated attribute is added either to 
	 * jsdl or posix. 
	 * @param attributeName
	 * @param attributeValue
	 * @param jsdl
	 * @param posix
	 */
	public void translateAttribute(AttrName attributeName, 
			Expr attributeValue, 
			RecordExpr jdl, 
			JobDescription_Type jsdl, 
			POSIXApplication_Type posix) throws Exception {
		
		String name = attributeName.toString(); 
		
		if("Executable".equalsIgnoreCase(name)) {
			FileName_Type executable = new FileName_Type(); 
			executable.set_value(attributeValue.stringValue()); 
			posix.setExecutable(executable); 
		}
		if("StdInput".equalsIgnoreCase(name)) {
			FileName_Type stdin = new FileName_Type(); 
			stdin.set_value(attributeValue.stringValue()); 
			posix.setInput(stdin); 
		}
		if("StdOutput".equalsIgnoreCase(name)) {
			FileName_Type stdout = new FileName_Type(); 
			stdout.set_value(attributeValue.stringValue()); 
			posix.setOutput(stdout); 
		}
		if("StdError".equalsIgnoreCase(name)) {
			FileName_Type stdout = new FileName_Type(); 
			FileName_Type stderr = new FileName_Type(); 
			stderr.set_value(attributeValue.stringValue()); 
			posix.setError(stderr); 
		}
		if("NodeNumber".equalsIgnoreCase(name)) {
			Resources_Type resource = jsdl.getResources();
			RangeValue_Type value = new RangeValue_Type(); 
			Exact_Type exactValue[] = new Exact_Type[]{new Exact_Type()}; 
			exactValue[0].set_value((double) attributeValue.intValue()); 
			value.setExact(exactValue); 
			resource.setTotalCPUCount(value); 
		}
		// "Arguments" is a space-separated string, 
		//  parse it using a stringtokenizer
		if("Arguments".equalsIgnoreCase(name)) {
			String args = attributeValue.stringValue(); 
		     StringTokenizer tokenizer = new StringTokenizer(args);
		     List argsList = new LinkedList(); 
		     while (tokenizer.hasMoreTokens()) {
		     	argsList.add(tokenizer.nextToken()); 
		     }
		     Argument_Type arguments[] = new Argument_Type[argsList.size()];
		     for(int i = 0; i < argsList.size(); i++) {
		    	Argument_Type argument = new Argument_Type(); 
				NormalizedString argString = new NormalizedString(); 
				argString.setValue((String)argsList.get(i)); 
				argument.set_value(argString); 
				arguments[i] = argument; 
		     }
			posix.setArgument(arguments); 
		}
		// Environment is specified as a list of strings, e.g.: 
		// {"HOME=/home/tordsson", "PI=3.14"}; 
		if("Environment".equalsIgnoreCase(name)) {
			ListExpr environmentList = (ListExpr) attributeValue; 
			Iterator iterator = environmentList.iterator(); 
			StringTokenizer tokenizer; 
			String envName; 
			String envValue;
			List envList = new LinkedList(); 
			while(iterator.hasNext()) {
				String envString = ((Expr)iterator.next()).stringValue(); 
				tokenizer = new StringTokenizer(envString, "="); 
				try { 
					envName = tokenizer.nextToken(); 
					envValue = tokenizer.nextToken(); 
				} catch (Exception e) {
					System.err.println("Invalid environment: " + envString); 
					throw e; 
				}
				Environment_Type env = new Environment_Type(); 
				env.setName(new NCName(envName)); 
				env.set_value(envValue); 
				envList.add(env); 
			}
			Environment_Type environment[] = new Environment_Type[envList.size()]; 
			for(int i = 0; i < envList.size(); i++) {
				environment[i] = (Environment_Type) envList.get(i); 
			}
			posix.setEnvironment(environment); 
		}
		/* Output files are specified as: 
		 * OutputSandbox = {"myoutput.txt", "myerror.txt" };
		 * OutputSandboxDestURI = "gsiftp://lxplus.cern.ch:3344/home/egtes";
		 */
		if("OutputSandbox".equalsIgnoreCase(name)) { 
			Expr sandboxDestURI = jdl.lookup("OutputSandboxDestURI"); 
			if(sandboxDestURI == null) {
				return; 
			}
			URI uri = new URI(sandboxDestURI.stringValue()); 
			ListExpr outputFiles = (ListExpr) attributeValue; 
			Iterator iterator = outputFiles.iterator(); 
			while(iterator.hasNext()) {
				String fileName = ((Expr)iterator.next()).stringValue(); 
				DataStaging_Type newTransfer = new DataStaging_Type(); 
				newTransfer.setFileName(fileName); 
				SourceTarget_Type target = new SourceTarget_Type(); 
				target.setURI(uri); 
				newTransfer.setTarget(target); 
				this.addTransfer(jsdl, newTransfer);
			}
		}
		// InputSandbox, format is the same as outputsandbox, that is: 
		// InputSandbox = {"data.in", "params"} 
		// The InputSandboxBaseURI attribute is optional, and specifies 
		// from where the files should be staged. Unless this attribute 
		// is specified, the input files are staged from the local host. 
		if("InputSandbox".equalsIgnoreCase(name)) {
			Expr sandboxBaseURI = jdl.lookup("InputSandboxBaseURI"); 
			
			// We need to make this translations at the client side, 
			// the the JSS does not know the CWD of the client invocing it. 
			URI uri; 
			if(sandboxBaseURI == null) {
				// Let uri be localhost
				InetAddress address = InetAddress.getLocalHost(); 
				String localHost = address.getCanonicalHostName(); 
				String URIString =  "gsiftp://"  + localHost + ":2811"; 
				uri = new URI(URIString); 
			}
			else {
				uri = new URI(sandboxBaseURI.stringValue()); 
			}
			
			File dummy = new File(""); 
			String path = dummy.getCanonicalPath(); 
			
			ListExpr inputFiles = (ListExpr) attributeValue; 
			Iterator iterator = inputFiles.iterator(); 
			// Add path = CWD to all files specified without path.
			while(iterator.hasNext()) {
				String fileName = ((Expr)iterator.next()).stringValue(); 
				if('/' != fileName.charAt(0)) { 					
					fileName = path + "/" + fileName; 
				}
				
				DataStaging_Type newTransfer = new DataStaging_Type(); 
				newTransfer.setFileName(fileName); 
				SourceTarget_Type source = new SourceTarget_Type(); 
				source.setURI(uri); 
				newTransfer.setSource(source); 
				this.addTransfer(jsdl, newTransfer); 				
			}		
		}
		// This is a free text format attribute, referring to 
		// glue properties of the selected resource. These properties
		// are expressed as other.GlueCEInfoTotalCPUs etc.
		if("Requirements".equalsIgnoreCase(name)) {
			this.translateRequirements(jsdl, attributeValue); 
		}
		
		// TODO: These extensions must be documented. 
		// N.B: These attributes are not defined, they are just a 
		// hack of mine. 
		if("CPUTime".equalsIgnoreCase(name)) {
			double time = (double) attributeValue.intValue(); 
			double timeInSeconds = time * 60; 
			Resources_Type resource = jsdl.getResources();
			RangeValue_Type value = new RangeValue_Type();
			Exact_Type exact[] = new Exact_Type[] {
					new Exact_Type(timeInSeconds)}; 
			value.setExact(exact); 
			resource.setIndividualCPUTime(value); 
			
		}
		if("WallTime".equalsIgnoreCase(name)) {
			double time = (double) attributeValue.intValue(); 
			double timeInSeconds = time * 60; 
			Resources_Type resource = jsdl.getResources();
			RangeValue_Type value = new RangeValue_Type();
			Exact_Type exact[] = new Exact_Type[] {
					new Exact_Type(timeInSeconds)}; 
			value.setExact(exact); 
			resource.setIndividualCPUTime(value); 
		}
	}
	
	public void validateJob(RecordExpr record) throws Exception {
		
		Expr type = record.lookup("Type"); 
		
		if(type == null || !"Job".equalsIgnoreCase(type.stringValue())) {
			System.err.println("JDL must be of type \"Job\""); 
			throw new Exception("JDL must be of type \"Job\""); 
		}
		Expr jobType = record.lookup("JobType"); 
		if(jobType != null && !"Normal".equalsIgnoreCase(jobType.stringValue())) {
			System.err.println("JDL must be of JobType \"Normal\""); 
			throw new Exception("JDL must be of JobType \"Normal\""); 
		}
	}
	
	public Object translate(Object jobDescriptionFile) throws Exception {
				
		// Create jsdl object: 
		JobDefinition_Type jobDefinition = new JobDefinition_Type(); 
		JobDescription_Type jobDescription = new JobDescription_Type(); 
		Application_Type application = new Application_Type(); 
		Resources_Type resource = new Resources_Type(); 
		jobDescription.setResources(resource); 
		POSIXApplication_Type posixApplication = new POSIXApplication_Type(); 

		// READ JDL from file: 
        FileReader input = new FileReader((String)jobDescriptionFile); 
        ClassAdParser parser = new ClassAdParser(input, ClassAdParser.TEXT); 
        parser.setVerbosity(0); // Turn off error messages
        
        Expr expr = parser.parse(); 
        if(expr == null) {
        	throw new Exception("Invalid job description (syntax error)"); 
        }
        
        while(expr != null) {
        	
            if(expr.type == Expr.LIST) {
            	throw new Exception("Job description is of invalid type"); 
            }
            if(expr.type == Expr.RECORD) {
            	RecordExpr record = (RecordExpr) expr; 
            	
            	// Ensure that type and jobType are correct. 
            	validateJob(record); 
            	
            	Iterator it = record.attributes(); 
            	while(it.hasNext()) {
					AttrName attrName = (AttrName) it.next(); 
            		Expr value = record.lookup(attrName);
            		translateAttribute(attrName, 
            				value, 
							record, 
							jobDescription, 
							posixApplication); 
            	}
            }
            expr = parser.parse(); 
        }
        
		// Connect jsdl object. 
		MessageElement any[] = AnyHelper.toAnyArray(posixApplication); 
		application.set_any(any); 
		jobDescription.setApplication(application); 
		jobDefinition.setJobDescription(jobDescription); 
		
        return jobDefinition; 
    }
}
