/*
 * Created on Sep 13, 2005
 *
 */
package se.umu.wsag.impl;

import java.io.File; 
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.List; 

import org.globus.gsi.GlobusCredential; 
import org.apache.log4j.Logger; 

/**
 * @author tordsson
 *
 * This class ensures that the proxy file used for creating 
 * the advance reservation is saved in the control dir 
 * under the name reservation.resID.proxy, enabling later
 * comparisons with the proxy used for job submission, 
 * thus ensuring that the user submitting the job is the 
 * one creating the reservation. 
 */
public class ARCLocalAgreementMapping implements LocalAgreementMapping {

	private static Logger logger = Logger.getLogger(
			ARCLocalAgreementMapping.class); 
	
	private String controlDir; 
	
	public void init(Object context) {		
		this.controlDir = (String) context; 		
		logger.debug("Using controlDir: " + controlDir); 
	}
	
	/**
	 * Stores proxy file in control dir in file 
	 * reservation.reservationID.proxy allowing the GM to 
	 * compare the proxy used to submit the job with the one 
	 * used to create the reservation. 
	 */
	public void create(Object context) throws Exception {
		
		List l = (List) context; 
		
		GlobusCredential delegatedCredential = (GlobusCredential) l.get(0); 
		String localID = (String) l.get(1); 
		
        // Store user proxy for later comparision. 
		// Typically, controlDir is /var/spool/nordugrid/jobstatus/
        String proxyPath = 
        	this.controlDir + "reservation." + ""+ localID + ".proxy"; 
        OutputStream proxyLocation = new FileOutputStream(proxyPath);         
        logger.debug("Saving proxy for user: " + delegatedCredential.
        		getIdentity() + " to file: " + proxyPath); 
        delegatedCredential.save(proxyLocation);  
		
	}
	
	/**
	 * Removes proxy created for reservation in ARC control dir. 
	 */
	public void cancel(Object context) throws Exception {
		
		String localID = (String) context; 
				 
		// Remove saved proxy file...
		String proxyPath = this.controlDir + "reservation." + localID + ".proxy"; 
		logger.debug("Removing proxy file:\n " + proxyPath); 
		File proxyFile = new File(proxyPath); 
		proxyFile.delete(); 
	
	}
}
